@@ -1,7 +1,7 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/wwwoffles.c 2.67.1.1 1998/07/10 15:00:39 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/wwwoffles.c 2.84 1998/09/03 17:32:05 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.2c.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.3b.
   A server to fetch the required pages.
   ******************/ /******************
   Written by Andrew M. Bishop
@@ -32,6 +32,7 @@
 #include "errors.h"
 
 
+static int request_new(URL *Url,char *referer);
 static void uninstall_sighandlers(void);
 
 
@@ -42,6 +43,7 @@ typedef enum _Mode
  Real,                          /*+ From server host to cache and client. +*/
  RealNoCache,                   /*+ From server host to client. +*/
  RealRefresh,                   /*+ Refresh the page, forced from index. +*/
+ RealPassword,                  /*+ From server host to cache, then again with a password. +*/
  SpoolOrReal,                   /*+ Spool if already cached else Real. +*/
  Fetch,                         /*+ From server host to cache. +*/
  Spool,                         /*+ From cache to client. +*/
@@ -69,14 +71,14 @@ int wwwoffles(int online,int browser,int client)
 {
  int outgoing=-1,spool=-1;
  int fetch_again=0;
- char *mirrorProto=NULL,*mirrorHost=NULL;
+ char *aliasProto=NULL,*aliasHost=NULL,*aliasPath=NULL;
  char *request_head,*request_body;
  char *reply_head,*reply_body;
  int reply_status=-1;
  char *url;
- URL *Url;
+ URL *Url,*Urlpw=NULL;
  Mode mode=None;
- int lasttime_exists=0,spool_exists=0;
+ int lasttime_exists=0,spool_exists=0,spool_exists_pw=0;
 
  /* Initialise things. */
 
@@ -134,34 +136,65 @@ int wwwoffles(int online,int browser,int client)
     exit(1);
    }
 
- PrintMessage(Inform,"URL='%s'.",url);
-
  Url=SplitURL(url);
 
- PrintMessage(Debug,"proto='%s'; host='%s'; path='%s'; args='%s'.",Url->proto,Url->host,Url->path,Url->args);
-
+ PrintMessage(Inform,"URL='%s'%s.",Url->name,Url->user?" (With username/password)":"");
+ PrintMessage(Debug,"proto='%s'; host='%s'; path='%s'; args='%s'; user:pass='%s:%s'.",
+              Url->proto,Url->host,Url->path,Url->args,Url->user,Url->pass);
  PrintMessage(ExtraDebug,"Incoming Request Head (from browser)\n%s",request_head);
 
- if(Url->user)
+ /* Check for an alias. */
+
+ if(IsAliased(Url->proto,Url->host,Url->path,&aliasProto,&aliasHost,&aliasPath))
    {
-    PrintMessage(Inform,"URLs with username and password are not allowed.",Url->proto);
-    if(mode==Real || mode==Spool || mode==SpoolOrReal)
-       HTMLMessage(client,404,"WWWOFFLE Password Not Allowed",NULL,"PasswordNotAllowed",
-                   "url",Url->name,
-                   NULL);
-    exit(1);
+    URL *newUrl;
+    char *newurl=(char*)malloc(strlen(aliasProto)+strlen(aliasHost)+strlen(aliasPath)-
+                               strlen(Url->path)+strlen(Url->pathp)+8);
+
+    sprintf(newurl,"%s://%s%s%s",aliasProto,aliasHost,aliasPath,Url->pathp+strlen(Url->path));
+
+    newUrl=SplitURL(newurl);
+
+    if(Url->user)
+       AddURLPassword(newUrl,Url->user,Url->pass);
+
+    url=newurl;
+    FreeURL(Url);
+    Url=newUrl;
+
+    PrintMessage(Inform,"Aliased URL='%s'%s.",newUrl->name,newUrl->user?" (With username/password)":"");
+    PrintMessage(Debug,"Aliased proto='%s'; host='%s'; path='%s'; args='%s'; user:pass='%s:%s'.",
+                 newUrl->proto,newUrl->host,newUrl->path,newUrl->args,newUrl->user,newUrl->pass);
    }
 
  /* Initial sanity checks. */
 
  if(IsNotGot(Url->proto,Url->host,Url->path))
    {
-    PrintMessage(Inform,"The server '%s://%s' and/or path '%s' is on the list not to get.",Url->proto,Url->host,Url->path);
-    if(mode==Real || mode==Spool || mode==SpoolOrReal)
-       HTMLMessage(client,404,"WWWOFFLE Host Not Got",NULL,"HostNotGot",
-                   "url",Url->name,
-                   NULL);
-    exit(0);
+    if(!DontGetReplacementURL)
+      {
+       PrintMessage(Inform,"The server '%s://%s' and/or path '%s' is on the list not to get.",Url->proto,Url->host,Url->path);
+       if(mode==Real || mode==Spool || mode==SpoolOrReal)
+          HTMLMessage(client,404,"WWWOFFLE Host Not Got",NULL,"HostNotGot",
+                      "url",Url->name,
+                      NULL);
+       exit(0);
+      }
+    else
+      {
+       char *newurl=(char*)malloc(strlen(DontGetReplacementURL)+1);
+       URL *newUrl=SplitURL(DontGetReplacementURL);
+
+       strcpy(newurl,DontGetReplacementURL);
+
+       url=newurl;
+       FreeURL(Url);
+       Url=newUrl;
+
+       PrintMessage(Inform,"Replaced URL='%s'%s.",newUrl->name,newUrl->user?" (With username/password)":"");
+       PrintMessage(Debug,"Replaced proto='%s'; host='%s'; path='%s'; args='%s'; user:pass='%s:%s'.",
+                    newUrl->proto,newUrl->host,newUrl->path,newUrl->args,newUrl->user,newUrl->pass);
+      }    
    }
 
  if(!Url->Protocol)
@@ -176,7 +209,8 @@ int wwwoffles(int online,int browser,int client)
    }
 
  /* Change the mode based on the URL as required. */
- /* (At this point mode is Spool, Real, SpoolOrReal or Fetch only.) */
+
+ /* mode = Spool, Real, SpoolOrReal or Fetch */
 
  /* A local URL. */
 
@@ -186,51 +220,83 @@ int wwwoffles(int online,int browser,int client)
       {
        int recurse=0;
 
-       url=RefreshPage(client,Url,request_body,&recurse);
+       char *newurl=RefreshPage(client,Url,request_body,&recurse);
 
-       if(!url)
+       if(!newurl)
           exit(0);
        else
          {
-          FreeURL(Url);
-          Url=SplitURL(url);
+          URL *newUrl=SplitURL(newurl);
 
-          if(Url->args && *Url->args=='!')
+          if(Url->user)
+             AddURLPassword(newUrl,Url->user,Url->pass);
+
+          if(newUrl->args && *newUrl->args=='!')
             {
              PrintMessage(Inform,"It is not possible to refresh a URL that was posted.");
              if(mode!=Fetch)
                 HTMLMessage(client,404,"WWWOFFLE Cant Refresh Posted",NULL,"CantRefreshPosted",
-                            "url",Url->name,
+                            "url",newUrl->name,
                             NULL);
              exit(0);
             }
 
-          if(recurse)
+          if(recurse && Url->args && *Url->args=='!')
             {
              if(mode==Spool || mode==Real || mode==SpoolOrReal)
                {
                 mode=SpoolGet;
                 free(request_head);
-                request_head=RequestURL(Url->name,NULL);
+                request_head=RequestURL(newUrl,NULL);
                 if(request_body)
                    free(request_body);
                 request_body=NULL;
                }
              fetch_again=1;
             }
+          else if(recurse)
+            {
+             if(mode==Fetch)
+                fetch_again=1;
+             else if(mode==Spool)
+               {
+                mode=SpoolGet;
+                free(newurl);
+                FreeURL(newUrl);
+                newurl=(char*)malloc(strlen(url)+1);
+                strcpy(newurl,url);
+                newUrl=SplitURL(newurl);
+                if(Url->user)
+                   AddURLPassword(newUrl,Url->user,Url->pass);
+               }
+             else if(mode==Real || mode==SpoolOrReal)
+               {
+                mode=SpoolGet;
+                free(request_head);
+                request_head=RequestURL(newUrl,NULL);
+                if(request_body)
+                   free(request_body);
+                request_body=NULL;
+               }
+            }
           else
             {
              if(mode==Spool)
                 mode=SpoolRefresh;
-             else if(mode==SpoolOrReal || mode==Real)
+             else
                {
-                DeleteWebpageSpoolFile(Url,0);
+                DeleteWebpageSpoolFile(newUrl,0);
                 mode=RealRefresh;
                }
             }
 
-          PrintMessage(Inform,"new URL='%s'.",url);
-          PrintMessage(Debug,"new proto='%s'; host='%s'; path='%s'; args='%s'.",Url->proto,Url->host,Url->path,Url->args);
+          url=newurl;
+          FreeURL(Url);
+          Url=newUrl;
+
+          PrintMessage(Inform,"New URL='%s'%s.",newUrl->name,newUrl->user?" (With username/password)":"");
+          PrintMessage(Debug,"New proto='%s'; host='%s'; path='%s'; args='%s'; user:pass='%s:%s'.",
+                       newUrl->proto,newUrl->host,newUrl->path,newUrl->args,newUrl->user,newUrl->pass);
          }
       }
     else if(mode==Fetch)
@@ -316,41 +382,82 @@ int wwwoffles(int online,int browser,int client)
  else if(mode==Fetch)
     ParseRecurseOptions(NULL);
 
- /* Check for a mirror site. */
+ /* mode = Spool, SpoolGet, SpoolRefresh, Real, RealRefresh, SpoolOrReal or Fetch */
 
- if(IsMirrored(Url->proto,Url->host,&mirrorProto,&mirrorHost))
+ /* Check for a username / password */
+
+ if(Url->user)
    {
-    char *newurl=(char*)malloc(strlen(mirrorProto)+strlen(mirrorHost)+strlen(Url->pathp)+8);
+    URL *new=SplitURL(Url->name);
 
-    sprintf(newurl,"%s://%s%s",mirrorProto,mirrorHost,Url->pathp);
+    Urlpw=Url;
+    Url=new;
 
-    url=newurl;
-
-    FreeURL(Url);
-    Url=SplitURL(url);
-
-    PrintMessage(Inform,"mirrored URL='%s'.",url);
-    PrintMessage(Debug,"mirrored proto='%s'; host='%s'; path='%s'; args='%s'.",Url->proto,Url->host,Url->path,Url->args);
-
-    if(!Url->Protocol)
+    if(!Urlpw->pass)
       {
-       PrintMessage(Inform,"The protocol '%s' is not available.",Url->proto);
-       if(mode==Real || mode==Spool || mode==SpoolOrReal)
-          HTMLMessage(client,404,"WWWOFFLE Illegal Protocol",NULL,"IllegalProtocol",
-                      "url",Url->name,
-                      "protocol",Url->proto,
+       if(mode==Fetch)
+         {
+          FreeURL(Urlpw);
+          Urlpw=NULL;
+         }
+       else
+         {
+          HTMLMessage(client,403,"WWWOFFLE Username Needs Password",NULL,"UserNeedsPass",
+                      "url",Urlpw->name,
+                      "user",Urlpw->user,
                       NULL);
-       exit(1);
+          exit(0);
+         }
       }
    }
 
+ /* Check for an existing cached version */
+
+ spool_exists=ExistsWebpageSpoolFile(Url);
+
+ if(Urlpw)
+   {
+    spool_exists_pw=ExistsWebpageSpoolFile(Urlpw);
+
+    if((mode==Spool || mode==SpoolGet || mode==SpoolOrReal) && spool_exists)
+      {
+       if(SpooledPageStatus(Url)==401)
+         {FreeURL(Url);Url=Urlpw;Urlpw=NULL;
+         spool_exists=spool_exists_pw;spool_exists_pw=0;}
+       else
+         {FreeURL(Urlpw);Urlpw=NULL;
+         spool_exists_pw=0;}
+      }
+    else if(mode==Spool || mode==SpoolGet || mode==SpoolOrReal) /* && ! spool_exists */
+      {
+       if(spool_exists_pw)
+          request_new(Url,NULL);
+       FreeURL(Url);Url=Urlpw;Urlpw=NULL;
+       spool_exists=spool_exists_pw;spool_exists_pw=0;
+      }
+    else if(mode==Fetch && spool_exists)
+      {FreeURL(Url);Url=Urlpw;Urlpw=NULL;
+      spool_exists=spool_exists_pw;spool_exists_pw=0;}
+    else if(mode==Real && spool_exists)
+      {
+       if(SpooledPageStatus(Url)==401)
+         {FreeURL(Url);Url=Urlpw;Urlpw=NULL;
+         spool_exists=spool_exists_pw;spool_exists_pw=0;}
+       else
+         {FreeURL(Urlpw);Urlpw=NULL;
+         spool_exists_pw=0;}
+      }
+   }
+
+realpassword:
+
  /* Check if it needs to be cached. */
 
  if(IsLocalNetHost(Url->host))
    {
     if(mode==Real || mode==Spool || mode==SpoolOrReal)
        mode=RealNoCache;
-    else if(mode==SpoolWillGet || mode==SpoolGet)
+    else if(mode==SpoolGet)
        ;
     else
       {
@@ -363,6 +470,12 @@ int wwwoffles(int online,int browser,int client)
     mode=RealNoCache;
    }
 
+ if(mode==RealNoCache && Urlpw)
+   {FreeURL(Url);Url=Urlpw;Urlpw=NULL;
+   spool_exists=spool_exists_pw;spool_exists_pw=0;}
+
+ /* mode = Spool, SpoolGet, SpoolRefresh, Real, RealRefresh, RealNoCache, SpoolOrReal or Fetch */
+
  /* Check if it was posted. */
 
  if(!strncasecmp(request_head,"POST",4))
@@ -419,9 +532,18 @@ int wwwoffles(int online,int browser,int client)
 
  /* Check if a refresh is needed based on pragma, request changes, autodial. */
 
- spool_exists=ExistsWebpageSpoolFile(Url);
-
- if(PragmaNoCache && GetHTTPHeader(request_head,"Pragma: no-cache"))
+ if(mode==Real && ExistsLockWebpageSpoolFile(Url))
+   {
+    mode=Spool;
+   }
+ else if(mode==Fetch && ExistsLockWebpageSpoolFile(Url))
+   {
+    PrintMessage(Debug,"Already fetching URL");
+    if(client!=-1)
+       write_formatted(client,"Fetching %s [Already fetching]\n",Url->name);
+    exit(0);
+   }
+ else if(PragmaNoCache && GetHTTPHeader(request_head,"Pragma: no-cache"))
    {
     if(mode==Spool || mode==SpoolGet)
       {
@@ -445,12 +567,12 @@ int wwwoffles(int online,int browser,int client)
     else
        mode=SpoolGet;
    }
- else if(mode==Fetch && spool_exists && !fetch_again)
+ else if(mode==Fetch && spool_exists)
    {
     spool=OpenWebpageSpoolFile(1,Url);
     init_buffer(spool);
 
-    if(RequestChanges(spool,&request_head)==1)
+    if(RequestChanged>=0 && RequestChanges(spool,&request_head)==1)
        TouchWebpageSpoolFile(Url);
     else
        exit(0);
@@ -477,6 +599,8 @@ int wwwoffles(int online,int browser,int client)
        mode=Real;
    }
 
+ /* mode = Spool, SpoolGet, SpoolWillGet, SpoolPragma, SpoolRefresh, Real, RealRefresh, RealNoCache or Fetch */
+
  /* Create the last time file. */
 
  if(mode==Real || mode==Fetch)
@@ -486,8 +610,10 @@ int wwwoffles(int online,int browser,int client)
 
  if(mode==Real || mode==Fetch)
    {
+    if(spool_exists)
+       CreateBackupWebpageSpoolFile(Url);
+
     spool=OpenWebpageSpoolFile(0,Url);
-    init_buffer(spool);         /* may be read if Real changes to Spool later. */
 
     if(spool==-1)
       {
@@ -499,8 +625,7 @@ int wwwoffles(int online,int browser,int client)
        exit(1);
       }
 
-    if(spool_exists)
-       CreateBackupWebpageSpoolFile(Url,spool);
+    CreateLockWebpageSpoolFile(Url);
    }
  else if(mode==Spool || mode==SpoolPragma)
    {
@@ -539,6 +664,13 @@ int wwwoffles(int online,int browser,int client)
    {
     char *err=(Url->Protocol->open)(Url);
 
+    if(err && ConnectRetry)
+      {
+       PrintMessage(Inform,"Waiting to try connection again.");
+       sleep(10);
+       err=(Url->Protocol->open)(Url);
+      }
+
     if(err)
       {
        if(mode==Fetch || mode==Real)
@@ -551,6 +683,7 @@ int wwwoffles(int online,int browser,int client)
                       "cache","yes",
                       "backup",spool_exists?"yes":"",
                       NULL);
+          DeleteLockWebpageSpoolFile(Url);
          }
        if(mode==Real || mode==RealNoCache)
           HTMLMessage(client,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
@@ -586,14 +719,22 @@ int wwwoffles(int online,int browser,int client)
                       "cache","yes",
                       "backup",spool_exists?"yes":"",
                       NULL);
+          DeleteLockWebpageSpoolFile(Url);
          }
-       if(mode==Real || mode==RealNoCache)
+       if(mode==Real)
           HTMLMessage(client,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
                       "url",Url->name,
                       "reason",err,
                       "cache","",
                       "backup",spool_exists?"yes":"",
                       NULL);
+       else if(mode==RealNoCache)
+          HTMLMessage(client,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
+                      "url",Url->name,
+                      "reason",err,
+                      "cache","",
+                      "backup","",
+                      NULL);
        else if(mode==Fetch && client!=-1)
           write_formatted(client,"Fetching %s [Server Connection Error]\n",Url->name);
        exit(1);
@@ -635,6 +776,7 @@ int wwwoffles(int online,int browser,int client)
                       "cache","yes",
                       "backup",spool_exists?"yes":"",
                       NULL);
+          DeleteLockWebpageSpoolFile(Url);
          }
        if(mode==Real || mode==RealNoCache)
           HTMLMessage(client,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
@@ -660,48 +802,43 @@ int wwwoffles(int online,int browser,int client)
        if(!new_url)
           PrintMessage(Warning,"Cannot parse the reply for the new location.");
        else
-         {
-          char *new_request=RequestURL(new_url,Url->name);
-          int new_outgoing=OpenOutgoingSpoolFile(0);
-
-          PrintMessage(Inform,"This URL has been moved to %s.",new_url);
-
-          if(new_outgoing==-1)
-             PrintMessage(Warning,"Cannot open the new outgoing request to write.");
-          else
-            {
-             URL *new_Url=SplitURL(new_url);
-             write_string(new_outgoing,new_request);
-             CloseOutgoingSpoolFile(new_outgoing,new_Url);
-             FreeURL(new_Url);
-             fetch_again++;
-            }
-
-          free(new_request);
-          free(new_url);
-         }
+          fetch_again+=RecurseFetchRelocation(Url,new_url);
       }
     else if(reply_status==304)
       {
        PrintMessage(Inform,"Server page is not newer than the one in cache.");
 
-       if((mode==Fetch || mode==Real) && !lasttime_exists)
-          DeleteLastTimeSpoolFile(Url);
+       if(mode==Fetch || mode==Real)
+         {
+          close(spool);
+          DeleteLockWebpageSpoolFile(Url);
+          RestoreBackupWebpageSpoolFile(Url);
+          if(!lasttime_exists)
+             DeleteLastTimeSpoolFile(Url);
+         }
        if(mode==Fetch)
          {
           if(client!=-1)
              write_formatted(client,"Fetching %s [Page Unchanged]\n",Url->name);
-          if(spool_exists)
-             DeleteBackupWebpageSpoolFile(Url);
           exit(0);
          }
        else if(mode==Real)
          {
+          spool=OpenWebpageSpoolFile(1,Url);
+          init_buffer(spool);
+
           mode=Spool;
-          if(spool_exists)
-             DeleteBackupWebpageSpoolFile(Url);
          }
       }
+    else if(mode==Fetch && reply_status==401 && Urlpw)
+      {
+       if(request_new(Urlpw,Url->name))
+          fetch_again++;
+      }
+    else if(mode==Real && reply_status==401 && Urlpw)
+      {
+       mode=RealPassword;
+      }
     else if(mode==Fetch && client!=-1)
        write_formatted(client,"Fetching %s\n",Url->name);
    }
@@ -710,6 +847,8 @@ int wwwoffles(int online,int browser,int client)
 
  reply_body=(char*)malloc(257);
 
+ /* mode = Spool, SpoolGet, SpoolWillGet, SpoolPragma, SpoolRefresh, Real, RealRefresh, RealNoCache, RealPassword or Fetch */
+
  /* Close the outgoing file if any. */
 
  if(outgoing>=0)
@@ -720,60 +859,62 @@ int wwwoffles(int online,int browser,int client)
        CloseOutgoingSpoolFile(outgoing,Url);
    }
 
- /* Process the request. */
+ /* The main body of the handling of the data. */
 
- if(mode==Real || mode==RealNoCache)
+ if(mode==Real || mode==RealNoCache || mode==RealPassword)
    {
     int n=strlen(reply_head),err=0;
 
-    if(mode==Real)
+    if(mode!=RealNoCache)
        write_data(spool,reply_head,n);
-    err=write_data(client,reply_head,n);
+    if(mode!=RealPassword)
+       err=write_data(client,reply_head,n);
 
     while(err!=-1 && (n=(Url->Protocol->readbody)(reply_body,256))>0)
       {
-       if(mode==Real)
+       if(mode!=RealNoCache)
           write_data(spool,reply_body,n);
-       err=write_data(client,reply_body,n);
+       if(mode!=RealPassword)
+          err=write_data(client,reply_body,n);
       }
 
-    if(mode==Real)
+    if(mode!=RealNoCache)
+      {
        if(err==-1 || n<0)
          {
           lseek(spool,0,SEEK_SET);
           ftruncate(spool,0);
-          if(err==-1)
-            {
-             PrintMessage(Warning,"Error writing to client [%!s]; client disconnected?.");
-             HTMLMessage(spool,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
-                         "url",Url->name,
-                         "reason","ClientClose",
-                         "cache","yes",
-                         "backup",spool_exists?"yes":"",
-                         NULL);
-             exit(1);
-            }
-          else
-            {
-             PrintMessage(Warning,"Timed out while reading from remote host.");
-             HTMLMessage(spool,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
-                         "url",Url->name,
-                         "reason","TimeoutTransfer",
-                         "cache","yes",
-                         "backup",spool_exists?"yes":"",
-                         NULL);
-             exit(1);
-            }
+         }
+
+       if(err==-1)
+         {
+          PrintMessage(Warning,"Error writing to client [%!s]; client disconnected?.");
+          HTMLMessage(spool,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
+                      "url",Url->name,
+                      "reason","ClientClose",
+                      "cache","yes",
+                      "backup",spool_exists?"yes":"",
+                      NULL);
+         }
+       else if(n<0)
+         {
+          PrintMessage(Warning,"Timed out while reading from remote host.");
+          HTMLMessage(spool,503,"WWWOFFLE Remote Host Error",NULL,"RemoteHostError",
+                      "url",Url->name,
+                      "reason","TimeoutTransfer",
+                      "cache","yes",
+                      "backup",spool_exists?"yes":"",
+                      NULL);
          }
        else
-         {
-          int tell=lseek(spool,0,SEEK_CUR);
-          if(tell!=-1)
-             ftruncate(spool,tell);
-
           if(spool_exists)
              DeleteBackupWebpageSpoolFile(Url);
-         }
+
+       DeleteLockWebpageSpoolFile(Url);
+
+       if(err==-1 || n<0)
+          exit(1);
+      }
    }
  else if(mode==Fetch)
    {
@@ -795,17 +936,15 @@ int wwwoffles(int online,int browser,int client)
                    "cache","yes",
                    "backup",spool_exists?"yes":"",
                    NULL);
-       exit(1);
       }
     else
-      {
-       int tell=lseek(spool,0,SEEK_CUR);
-       if(tell!=-1)
-          ftruncate(spool,tell);
-
        if(spool_exists)
           DeleteBackupWebpageSpoolFile(Url);
-      }
+
+    DeleteLockWebpageSpoolFile(Url);
+
+    if(n<0)
+       exit(1);
 
     if(reply_status>=200 && reply_status<400)
       {
@@ -813,14 +952,35 @@ int wwwoffles(int online,int browser,int client)
        init_buffer(spool);
 
        if(ParseHTML(spool,Url,0))
-          fetch_again=RecurseFetch(Url);
+          fetch_again+=RecurseFetch(Url);
       }
    }
  else if(mode==Spool || mode==SpoolPragma)
    {
     struct stat buf;
     char *head="HTTP/1.0 503 WWWOFFLE Remote Host Error\r\n";
-    int n=read_data(spool,reply_body,64);
+    int n;
+
+    if(ExistsLockWebpageSpoolFile(Url))
+      {
+       int t=SocketTimeout;
+
+       PrintMessage(Inform,"Waiting for the page to be unlocked.");
+
+       while((t-=5) && ExistsLockWebpageSpoolFile(Url))
+          sleep(5);
+
+       if(t==0)
+         {
+          PrintMessage(Inform,"Timed out waiting for the page to be unlocked.");
+          HTMLMessage(spool,503,"WWWOFFLE File Locked",NULL,"FileLocked",
+                      "url",Url->name,
+                      NULL);
+          exit(0);
+         }
+      }
+
+    n=read_data(spool,reply_body,64);
 
     if(!strncmp(reply_body,head,strlen(head)) || (!fstat(spool,&buf) && buf.st_size==0))
       {
@@ -899,9 +1059,13 @@ int wwwoffles(int online,int browser,int client)
    }
  else if(mode==SpoolGet)
    {
-    if(OfflineRequests || fetch_again)
+    if(OfflineRequests)
       {
-       if(fetch_again)
+       if(ConfirmRequests)
+          HTMLMessage(client,200,"WWWOFFLE Confirm Request",NULL,"ConfirmRequest",
+                      "url",Url->name,
+                      NULL);
+       else if(fetch_again)
           HTMLMessage(client,404,"WWWOFFLE Refresh Will Get",NULL,"RefreshWillGet",
                       "url",Url->name,
                       NULL);
@@ -935,15 +1099,23 @@ int wwwoffles(int online,int browser,int client)
 
  /* Close down and exit. */
 
- if(mode==Real && ExistsOutgoingSpoolFile(Url))
+ if((mode==Real || mode==RealPassword) && ExistsOutgoingSpoolFile(Url))
     DeleteOutgoingSpoolFile(Url);
 
- if(mode==Real || mode==RealNoCache || mode==Fetch)
+ if(mode==Real || mode==RealNoCache || mode==RealPassword || mode==Fetch)
     (Url->Protocol->close)();
 
  if(spool>=0)
     close(spool);
 
+ if(mode==RealPassword)
+   {
+    FreeURL(Url);Url=Urlpw;Urlpw=NULL;
+    spool_exists=spool_exists_pw;spool_exists_pw=0;
+    mode=Real;
+    goto realpassword;
+   }
+
  if(request_head)
     free(request_head);
  if(request_body)
@@ -964,6 +1136,37 @@ int wwwoffles(int online,int browser,int client)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Make a new request.
+
+  int request_new Returns 1 if successful.
+
+  URL *Url The new URL to get.
+
+  char *referer The refering URL.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static int request_new(URL *Url,char *referer)
+{
+ char *new_request=RequestURL(Url,referer);
+ int new_outgoing=OpenOutgoingSpoolFile(0);
+ int retval=0;
+
+ if(new_outgoing==-1)
+    PrintMessage(Warning,"Cannot open the new outgoing request to write.");
+ else
+   {
+    write_string(new_outgoing,new_request);
+    CloseOutgoingSpoolFile(new_outgoing,Url);
+    retval=1;
+   }
+
+ free(new_request);
+
+ return(retval);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Uninstall the signal handlers.
   ++++++++++++++++++++++++++++++++++++++*/
